/*
 -------------------------------------------------------------------------------
 | Producer - Copyright (C) 2001-2007  Don Burns                               |
 |                                                                             |
 | This library is free software; you can redistribute it and/or modify        |
 | it under the terms of the GNU Lesser General Public License as published    |
 | by the Free Software Foundation; either version 3 of the License, or        |
 | (at your option) any later version.                                         |
 |                                                                             |
 | This library is distributed in the hope that it will be useful, but         |
 | WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  |
 | or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public     |
 | License for more details.                                                   |
 |                                                                             |
 | You should have received a copy of the GNU Lesser General Public License    |
 | along with this software; if not, write to the Free Software Foundation,    |
 | Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.               |
 |                                                                             |
 -------------------------------------------------------------------------------
 */


#include <Producer/Camera.h>
#include <Producer/Trackball.h>

#include <ProducerOSG/OsgSceneHandler.h>

#include "MyKeyboardMouseResponder.h"

int main( int argc, char **argv )
{
    // RenderSurface
    // -------------
    // Create the Render Surface - In this case a Window at x,y = 100,100 width,height=640,480
    Producer::ref_ptr<Producer::RenderSurface> renderSurface = new Producer::RenderSurface;
    renderSurface->setWindowRectangle( 100, 100, 640, 480 );
    renderSurface->setWindowName( "Simple Producer Based OpenSceneGraph Viewer Example" );

    // Create a Camera to View the Scene and provide it with the RenderSurface
    Producer::ref_ptr<Producer::Camera> camera = new Producer::Camera;
    camera->setRenderSurface( renderSurface.get() );

    // OpenSceneGraph integration with OsgSceneHandler
    // ----------------------------------------------
    // Note we are passing the command line arguments, so osg models and relevant 
    // command line arguments are passed to OSG.
    Producer::ref_ptr<Producer::OsgSceneHandler> osgSceneHandler = new Producer::OsgSceneHandler(argc, argv );
    camera->setSceneHandler( osgSceneHandler.get() );

    // Find out the bounds of the scene
    Producer::Vec3 sceneCenter;
    double sceneRadius;
    osgSceneHandler->getSceneBounds( sceneCenter, sceneRadius );

    // InputDevice
    // -----------
    // Create a KeyboardMouse
    Producer::ref_ptr<Producer::KeyboardMouse> keyboardMouse = new Producer::KeyboardMouse( renderSurface.get() );

    // Tell it what its callback is, and kick off a background thread to do 
    // Keybard amd Mouse processing.  The done flag is passed by reference 
    // and will be set when the ESC key is pressed
    bool done = false;
    Producer::ref_ptr<MyKeyboardMouseResponder> kbmres = new MyKeyboardMouseResponder(done);
    keyboardMouse->addResponder( kbmres.get() );
    keyboardMouse->startThread();

    // Manipulator
    // -----------
    // Create the trackball manipulator and ...
    Producer::ref_ptr<Producer::Trackball> trackball = new Producer::Trackball;

    //      ...enable a 'Z' up orientation,
    trackball->setOrientation( Producer::Trackball::Z_UP );

    //      ...Center the camera on the scene,
    trackball->translate(-sceneCenter[0],-sceneCenter[1],-sceneCenter[2]);

    //      ...Move it back the size of the scene's radius times three,
    trackball->setDistance(sceneRadius*3.0f);

    //      ... Set the reference point.  This is the position where the trackball 
    //         will return to when we call trackball->reset();
    trackball->setReference();

    // 'done' will be set by our Keyboard and Mouse callback when we press the 'ESC' key.
    while( !done )
    {
        // Call update on the scene
        osgSceneHandler->update();

        // Toggle backface culling on an 'c' key press
        if( kbmres->toggleBackfaceCulling() )
            osgSceneHandler->toggleBackfaceCulling();

        // Toggle lighing on an 'l' key press
        if( kbmres->toggleLighting() )
            osgSceneHandler->toggleLighting();

        // Toggle texturing on an 't' key press
        if( kbmres->toggleTexturing() )
            osgSceneHandler->toggleTexturing();

        // Toggle wireframe on an 'w' key press
        if( kbmres->toggleWireframe() )
            osgSceneHandler->toggleWireframe();

        // Reset the Trackball if the user presses the <space bar>
        if( kbmres->resetTrackball() )
            trackball->reset();

        // Update the view, by updating the Trackball with the most recent
        // mouse X, Y, and button state 
        trackball->input( kbmres->mx(), kbmres->my(), kbmres->mbutton() );

        // Camera View Control
        // ------------
        // Set the camera's view by the trackball's matrix
        camera->setViewByMatrix( trackball->getMatrix() );

        // Call the next cull/dispatch frame
        camera->frame();
    }
}
