/*
 -------------------------------------------------------------------------------
 | Producer - Copyright (C) 2001-2007  Don Burns                               |
 |                                                                             |
 | This library is free software; you can redistribute it and/or modify        |
 | it under the terms of the GNU Lesser General Public License as published    |
 | by the Free Software Foundation; either version 3 of the License, or        |
 | (at your option) any later version.                                         |
 |                                                                             |
 | This library is distributed in the hope that it will be useful, but         |
 | WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  |
 | or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public     |
 | License for more details.                                                   |
 |                                                                             |
 | You should have received a copy of the GNU Lesser General Public License    |
 | along with this software; if not, write to the Free Software Foundation,    |
 | Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.               |
 |                                                                             |
 -------------------------------------------------------------------------------
 */

#include <ProducerOSG/OsgSceneHandler.h>

#include <Producer/Math.h>

#include <osgDB/ReadFile>
#include <osgDB/WriteFile>

#include <osg/ArgumentParser>
#include <osg/BoundingSphere>
#include <osg/State>
#include <osg/StateSet>
#include <osg/StateAttribute>
#include <osg/LightModel>
#include <osg/TexEnv>
#include <osg/Material>
#include <osg/PolygonOffset>
#include <osg/Image>

#include <osg/Vec4>

#include <osgUtil/UpdateVisitor>
#include <osgUtil/GLObjectsVisitor>

#include <osg/Node>

#include <osg/Timer>
#include <osg/Notify>
#include <osg/Texture>
#include <osg/VertexProgram>
#include <osg/FragmentProgram>
#include <osg/AlphaFunc>
#include <osg/ColorMatrix>
#include <osg/CollectOccludersVisitor>
#include <osg/Shader>
#include <osg/ApplicationUsage>

#include <osg/GLU>

#include <osg/CullStack>


namespace Producer {

OsgSceneHandler::OsgSceneHandler( unsigned int options)
{
    _rootNode = new osg::Group;
    _initScribeSwitch();
    _setDefaults( options );
    _initCalled = false;
}

OsgSceneHandler::OsgSceneHandler( int argc, char **argv, unsigned int options )
{
    osg::ArgumentParser arguments( &argc, argv );
    _constructorImplementation( arguments, options );
}

OsgSceneHandler::OsgSceneHandler( osg::ArgumentParser &arguments, unsigned int options )
{
    _constructorImplementation( arguments, options );
}

void OsgSceneHandler::_constructorImplementation( osg::ArgumentParser &arguments, unsigned int options )
{
    _rootNode = new osg::Group;
    _initScribeSwitch();
    osg::ref_ptr<osg::Node> db = osgDB::readNodeFiles( arguments ); 
    if( db.valid() )
    {
        _scene = db.get();
        _rootNode->addChild( _scene.get() );
        _scribeSwitch->addChild( _scene.get() );
        //_rootNode->addChild( _scribeSwitch.get() );
    }

    _setDefaults( options );
    _initCalled = false;
}

osg::State *OsgSceneHandler::getState()
{
#if defined(OSG_RENDERINFO)
    if (!_renderInfo.getState())
#else
    if (!_state.valid() )
#endif
    {
        osg::notify(osg::INFO) << "Warning: no valid Producer::OsgSceneHandler::_state attached, creating a default state automatically."<< std::endl; 

        // note the constructor for osg::State will set ContextID to 0 
        // which will be fine to single context graphics
        // applications which is ok for most apps, but not multiple 
        // context/pipe applications.

#if defined(OSG_RENDERINFO)
        _renderInfo.setState(new osg::State);
#else
        _state = new osg::State;
#endif

    }
#if defined(OSG_RENDERINFO)
    return _renderInfo.getState();
#else
    return _state.get();
#endif

}

void *OsgSceneHandler::getScene()
{
    return (void *)getRootNode();
}

bool OsgSceneHandler::loadScene( const std::string &fileName, bool replace )
{
    osg::ref_ptr<osg::Node> newScene = osgDB::readNodeFile( fileName );
    if( newScene.valid() )
    {
        if( _scene.valid() && replace )
        {
            _rootNode->replaceChild( _scene.get(), newScene.get() );
            _scribeSwitch->replaceChild( _scene.get(), newScene.get() );
        }
        else
        {
            _rootNode->addChild( newScene.get() );
            _scribeSwitch->addChild( newScene.get() );
        }
        _scene = newScene.get();
        return true;
    }
    return false;
}

bool OsgSceneHandler::setScene( void *newScenePtr, bool replace )
{
    osg::ref_ptr<osg::Node> newScene = static_cast<osg::Node *>(newScenePtr);
    if( newScene.valid() )
    {
        if( _scene.valid() && replace )
        {
            _rootNode->replaceChild( _scene.get(), newScene.get() );
            _scribeSwitch->replaceChild( _scene.get(), newScene.get() );
        }
        else
        {
            _rootNode->addChild( newScene.get() );
            _scribeSwitch->addChild( newScene.get() );
        }
        _scene = newScene.get();
        _scene->getBound();
        return true;
    }
    return false;
}

bool OsgSceneHandler::getSceneBounds( Vec3 &center, double &radius )
{
    if( _scene.valid() )
    {
        osg::BoundingSphere bs = _scene->getBound();
        center.set( bs.center().ptr() );
        if( (radius = bs.radius()) < 0 )
            return false;
        return true;
    }
    return false;
}

osg::Node *OsgSceneHandler::getRootNode()
{ 
    return _scene.get(); 
}

void OsgSceneHandler::update()
{
    _frameStamp->setFrameNumber( _frameStamp->getFrameNumber() + 1 );
    _frameStamp->setReferenceTime( _timer->delta_s( _t0, _timer->tick() ));

    if(_updateVisitor.valid())
    {
        _updateVisitor->reset();

        _updateVisitor->setFrameStamp(_frameStamp.get());

        // use the frame number for the traversal number.
        if (_frameStamp.valid())
        {
             _updateVisitor->setTraversalNumber(_frameStamp->getFrameNumber());
        }

        if( _rootNode.valid() )
            _rootNode->accept(*_updateVisitor.get());

    }
    // now force a recompute of the bounding volume while we are still in
    // the read/write app phase, this should prevent the need to recompute
    // the bounding volumes from within the cull traversal which may be
    // multi-threaded.
    if( _rootNode.valid() )
        _rootNode->getBound();
}

void OsgSceneHandler::cull( Producer::Camera &camera)
{
    osg::State* state = getState();

    if (!_localStateSet)
    {
        _localStateSet = new osg::StateSet;
    }

    // Set the viewport
    int x, y;
    unsigned int width,height;
    camera.getProjectionRectangle( x, y, width, height );
    setViewport( x, y, width, height );

    state->setFrameStamp(_frameStamp.get());

    if(camera.getStereoMode() == Producer::Camera::StereoSettings::Mono )
        _cullMono( camera );
    else
        _cullStereo( camera );
}

void OsgSceneHandler::_cullStereo( Producer::Camera &camera)
{
    //////// RIGHT
    if (!_cullVisitorRight)
    {
        osg::notify(osg::INFO) << "Warning: no valid Producer::OsgSceneHandler::_cullVisitor attached, "
            "creating a default CullVisitor automatically."<< std::endl;
        _cullVisitorRight = new osgUtil::CullVisitor;

    }

    if (!_rendergraphRight)
    {
        osg::notify(osg::INFO) << "Warning: no valid Producer::OsgSceneHandler::_rendergraph "
            "attached, creating a global default StateGraph automatically."<< std::endl;
        _rendergraphRight = new osgUtil::StateGraph;
    }

    if (!_renderStageRight)
    {
        osg::notify(osg::INFO) << 
            "Warning: no valid Producer::OsgSceneHandler::_renderStage attached, "
            "creating a default RenderStage automatically."<< std::endl;

        _renderStageRight = new OsgSceneHandler::RenderStage;
    }

    if( _cullVisitorRight.valid() )
    {
        _cullVisitorRight->setTraversalMask(_cullMask);
        camera.getLens()->setStereoEye( Producer::Camera::Lens::Right );
        _cullStage(  osg::Matrixd(camera.getProjectionMatrix()),
                     osg::Matrixd(camera.getViewMatrix()),
                    _cullVisitorRight.get(),
                    _rendergraphRight.get(),
                    _renderStageRight.get());

    }

    //////// LEFT
    if (!_cullVisitorLeft)
    {
        osg::notify(osg::INFO) << "Warning: no valid Producer::OsgSceneHandler::_cullVisitor attached, "
            "creating a default CullVisitor automatically."<< std::endl;
        _cullVisitorLeft = new osgUtil::CullVisitor;

    }

    if (!_rendergraphLeft)
    {
        osg::notify(osg::INFO) << "Warning: no valid Producer::OsgSceneHandler::_rendergraph "
            "attached, creating a global default StateGraph automatically."<< std::endl;
        _rendergraphLeft = new osgUtil::StateGraph;
    }

    if (!_renderStageLeft)
    {
        osg::notify(osg::INFO) << 
            "Warning: no valid Producer::OsgSceneHandler::_renderStage attached, "
            "creating a default RenderStage automatically."<< std::endl;

        _renderStageLeft = new OsgSceneHandler::RenderStage;
    }

    if( _cullVisitorLeft.valid() )
    {
        _cullVisitorLeft->setTraversalMask(_cullMask);
        camera.getLens()->setStereoEye( Producer::Camera::Lens::Left );
        _cullStage(  osg::Matrixd(camera.getProjectionMatrix()),
                     osg::Matrixd(camera.getViewMatrix()),
                    _cullVisitorLeft.get(),
                    _rendergraphLeft.get(),
                    _renderStageLeft.get());

    }
}

void OsgSceneHandler::_cullMono( Producer::Camera &camera)
{
    if (!_cullVisitor)
    {
        osg::notify(osg::INFO) << "Warning: no valid Producer::OsgSceneHandler::_cullVisitor attached, "
            "creating a default CullVisitor automatically."<< std::endl;
        _cullVisitor = new osgUtil::CullVisitor;

    }

    if (!_rendergraph)
    {
        osg::notify(osg::INFO) << "Warning: no valid Producer::OsgSceneHandler::_rendergraph "
            "attached, creating a global default StateGraph automatically."<< std::endl;
        _rendergraph = new osgUtil::StateGraph;
    }

    if (!_renderStage)
    {
        osg::notify(osg::INFO) << 
            "Warning: no valid Producer::OsgSceneHandler::_renderStage attached, "
            "creating a default RenderStage automatically."<< std::endl;

        _renderStage = new OsgSceneHandler::RenderStage;
    }

    if( _cullVisitor.valid() )
    {
        _cullVisitor->setTraversalMask(_cullMask);
        _cullStage(  osg::Matrixd(camera.getProjectionMatrix()),
                     osg::Matrixd(camera.getViewMatrix()),
                    _cullVisitor.get(),
                    _rendergraph.get(),
                    _renderStage.get());

    }
}


void OsgSceneHandler::draw( Producer::Camera & camera)
{
    if(camera.getStereoMode() == Producer::Camera::StereoSettings::Mono )
        _drawMono( camera );
    else
        _drawStereo( camera );
}

void OsgSceneHandler::_drawStereo( Producer::Camera & camera)
{
    osg::State* state = getState();

    // First eye
    if( camera.getFrameCount() != _stereoFrame )
    {
        if (_requiresFlush)
        {
            double availableTime = 0.005;
            flushDeletedGLObjects(availableTime);
        }

        // assume the the draw which is about to happen could generate GL objects that need flushing in the next frame.
        _requiresFlush = true;
        getState()->setInitialViewMatrix(new osg::RefMatrix(osg::Matrix(camera.getViewMatrix())));
        _localStateSet->setAttribute(getViewport());
    }

    osgUtil::RenderLeaf* previous = NULL;

    switch( camera.getLens()->getStereoEye() )
    {
        case Camera::Lens::Left :
#if defined(OSG_RENDERINFO)
            _renderStageLeft->drawPreRenderStages(_renderInfo,previous);
            _renderStageLeft->draw(_renderInfo,previous);
#else
            _renderStageLeft->drawPreRenderStages(*_state.get(),previous);
            _renderStageLeft->draw(*_state.get(),previous);
#endif
            break;

        case Camera::Lens::Right :
#if defined(OSG_RENDERINFO)
            _renderStageRight->drawPreRenderStages(_renderInfo,previous);
            _renderStageRight->draw(_renderInfo,previous);
#else
            _renderStageRight->drawPreRenderStages(*_state.get(),previous);
            _renderStageRight->draw(*_state.get(),previous);
#endif
            break;
    }


    // re apply the defalt OGL state.
    state->popAllStateSets();
    state->apply();

    // Second eye
    if( camera.getFrameCount() == _stereoFrame )
    {
        if (state->getCheckForGLErrors()!=osg::State::NEVER_CHECK_GL_ERRORS)
        {
            GLenum errorNo = glGetError();
            if (errorNo!=GL_NO_ERROR)
            {
            osg::notify(osg::WARN)<<"Warning: detected OpenGL error '"<<gluErrorString(errorNo)<<"'"<< std::endl;

            // go into debug mode of OGL error in a fine grained way to help
            // track down OpenGL errors.
            state->setCheckForGLErrors(osg::State::ONCE_PER_ATTRIBUTE);
            }
        }
    }
    _stereoFrame = camera.getFrameCount();
}

void OsgSceneHandler::_drawMono( Producer::Camera & camera)
{
    // note, to support multi-pipe systems the deletion of OpenGL display list
    // and texture objects is deferred until the OpenGL context is the correct
    // context for when the object were originally created.  Here we know what
    // context we are in so can flush the appropriate caches.
    if (_requiresFlush)
    {
        double availableTime = 0.005;
        flushDeletedGLObjects(availableTime);
    }

    // assume the the draw which is about to happen could generate GL objects that need flushing in the next frame.
    _requiresFlush = true;
    osg::State* state = getState();
    state->setInitialViewMatrix(new osg::RefMatrix(osg::Matrix(camera.getViewMatrix())));
    _localStateSet->setAttribute(getViewport());

    osgUtil::RenderLeaf* previous = NULL;

#if defined(OSG_RENDERINFO)
    _renderStage->drawPreRenderStages(_renderInfo,previous);
    _renderStage->draw(_renderInfo,previous);
#else
    _renderStage->drawPreRenderStages(*_state.get(),previous);
    _renderStage->draw(*_state.get(),previous);
#endif

    // re apply the defalt OGL state.
    state->popAllStateSets();
    state->apply();

    if (state->getCheckForGLErrors()!=osg::State::NEVER_CHECK_GL_ERRORS)
    {
        GLenum errorNo = glGetError();
        if (errorNo!=GL_NO_ERROR)
        {
            osg::notify(osg::WARN)<<"Warning: detected OpenGL error '"<<gluErrorString(errorNo)<<"'"<< std::endl;

            // go into debug mode of OGL error in a fine grained way to help
            // track down OpenGL errors.
            state->setCheckForGLErrors(osg::State::ONCE_PER_ATTRIBUTE);
        }
    }
}

void OsgSceneHandler::flushDeletedGLObjects(double& availableTime)
{
    osg::State* state = getState();

    if( state == 0L )
        return;

    _requiresFlush = false;

    double currentTime = state->getFrameStamp()?state->getFrameStamp()->getReferenceTime():0.0;

    osg::FrameBufferObject::flushDeletedFrameBufferObjects(state->getContextID(),currentTime,availableTime);
    osg::RenderBuffer::flushDeletedRenderBuffers(state->getContextID(),currentTime,availableTime);
    osg::Texture::flushDeletedTextureObjects(state->getContextID(),currentTime,availableTime);
    osg::Drawable::flushDeletedDisplayLists(state->getContextID(),availableTime);
    osg::Drawable::flushDeletedVertexBufferObjects(state->getContextID(),currentTime,availableTime);
    osg::VertexProgram::flushDeletedVertexProgramObjects(state->getContextID(),currentTime,availableTime);
    osg::FragmentProgram::flushDeletedFragmentProgramObjects(state->getContextID(),currentTime,availableTime);
    osg::Program::flushDeletedGlPrograms(state->getContextID(),currentTime,availableTime);
    osg::Shader::flushDeletedGlShaders(state->getContextID(),currentTime,availableTime);
}


void OsgSceneHandler::_setDefaults(unsigned int options)
{
    _frameStamp = new osg::FrameStamp;
    _timer = Producer::Timer::instance();
    _t0 = _timer->tick();

    _globalStateSet = new osg::StateSet;

    if ((options & HEADLIGHT) || (options & SKY_LIGHT))
    {
        _lightingMode=(options&HEADLIGHT) ? HEADLIGHT : SKY_LIGHT;
        _light = new osg::Light;
        _light->setLightNum(0);
        _light->setAmbient(osg::Vec4(0.00f,0.0f,0.00f,1.0f));
        _light->setDiffuse(osg::Vec4(0.8f,0.8f,0.8f,1.0f));
        _light->setSpecular(osg::Vec4(1.0f,1.0f,1.0f,1.0f));

        _globalStateSet->setAssociatedModes(_light.get(),osg::StateAttribute::ON);

        osg::LightModel* lightmodel = new osg::LightModel;
        lightmodel->setAmbientIntensity(osg::Vec4(0.1f,0.1f,0.1f,1.0f));
        _globalStateSet->setAttributeAndModes(lightmodel, osg::StateAttribute::ON);

        // enable lighting by default.
        _globalStateSet->setMode(GL_LIGHTING, osg::StateAttribute::ON);

    }
    else
    {
        _lightingMode = NO_SCENEVIEW_LIGHT;
    }
 
    getState();

    if (options & AUTOMATIC_CONTEXT_ID)
        getState()->setContextID(_contextIDCounter++);
    
    _rendergraph = new osgUtil::StateGraph;
    _renderStage = new OsgSceneHandler::RenderStage;


    if (options & COMPILE_GLOBJECTS_AT_INIT)
    {
        osgUtil::GLObjectsVisitor::Mode  dlvMode = 
                                            osgUtil::GLObjectsVisitor::COMPILE_DISPLAY_LISTS |
                                            osgUtil::GLObjectsVisitor::COMPILE_STATE_ATTRIBUTES | 
                                            osgUtil::GLObjectsVisitor::CHECK_BLACK_LISTED_MODES;

    #ifdef __sgi
        dlvMode = osgUtil::GLObjectsVisitor::COMPILE_STATE_ATTRIBUTES;
    #endif

        // sgi's IR graphics has a problem with lighting and display lists, as it seems to store 
        // lighting state with the display list, and the display list visitor doesn't currently apply
        // state before creating display lists. So will disable the init visitor default, this won't
        // affect functionality since the display lists will be created as and when needed.
        osgUtil::GLObjectsVisitor* dlv = new osgUtil::GLObjectsVisitor(dlvMode);
        dlv->setNodeMaskOverride(0xffffffff);
        _initVisitor = dlv;

    }
    
    _updateVisitor = new osgUtil::UpdateVisitor;

    _cullVisitor = new osgUtil::CullVisitor;

    _cullVisitor->setStateGraph(_rendergraph.get());
    _cullVisitor->setRenderStage(_renderStage.get());

    _globalStateSet->setGlobalDefaults();
    
    
    // enable depth testing by default.
    _globalStateSet->setMode(GL_DEPTH_TEST, osg::StateAttribute::ON);

    // set up an texture environment by default to speed up blending operations.
     osg::TexEnv* texenv = new osg::TexEnv;
     texenv->setMode(osg::TexEnv::MODULATE);
     _globalStateSet->setTextureAttributeAndModes(0,texenv, osg::StateAttribute::ON);

     _lighting = true;
     _texturing = true;
     _polygonMode = osg::PolygonMode::FILL;
     _scribe = false;
     _backfaceCulling = false; // Let's remain consistent with OSG.

     _stereoFrame = 0;
}

void OsgSceneHandler::_cullStage(
        const osg::Matrixd& projection,
        const osg::Matrixd& modelview,
        osgUtil::CullVisitor* cullVisitor, 
        osgUtil::StateGraph* rendergraph, 
        OsgSceneHandler::RenderStage *renderStage )
{
    if( !getViewport() )
        return;

    if (!_initCalled) _init();

    osg::ref_ptr<osg::RefMatrix> proj = new osg::RefMatrix(projection);
    osg::ref_ptr<osg::RefMatrix> mv = new osg::RefMatrix(modelview);

    // collect any occluder in the view frustum.
    if (_rootNode->containsOccluderNodes())
    {
        //std::cout << "Scene graph contains occluder nodes, searching for them"<<std::endl;
        
        if (!_collectOccludersVisitor) 
            _collectOccludersVisitor = new osg::CollectOccludersVisitor;
        
        _collectOccludersVisitor->inheritCullSettings(*this);
        
        _collectOccludersVisitor->reset();
        
        _collectOccludersVisitor->setFrameStamp(_frameStamp.get());

        // use the frame number for the traversal number.
        if (_frameStamp.valid())
        {
             _collectOccludersVisitor->setTraversalNumber(_frameStamp->getFrameNumber());
        }

        _collectOccludersVisitor->pushViewport(getViewport());
        _collectOccludersVisitor->pushProjectionMatrix(proj.get());
        _collectOccludersVisitor->pushModelViewMatrix(mv.get());

        // traverse the scene graph to search for occluder in there new positions.
        for(unsigned int i=0; i< _rootNode->getNumChildren(); ++i)
        {
            _rootNode->getChild(i)->accept(*_collectOccludersVisitor);
        }

        _collectOccludersVisitor->popModelViewMatrix();
        _collectOccludersVisitor->popProjectionMatrix();
        _collectOccludersVisitor->popViewport();
        
        // sort the occluder from largest occluder volume to smallest.
        _collectOccludersVisitor->removeOccludedOccluders();
        
        
        osg::notify(osg::DEBUG_INFO) << 
            "finished searching for occluder - found "<<
                _collectOccludersVisitor->getCollectedOccluderSet().size()<<std::endl;
           
        cullVisitor->getOccluderList().clear();
        std::copy(_collectOccludersVisitor->getCollectedOccluderSet().begin(),
                _collectOccludersVisitor->getCollectedOccluderSet().end(), 
                std::back_insert_iterator<osg::CullStack::OccluderList>(cullVisitor->getOccluderList()));
    }
    


    cullVisitor->reset();

    cullVisitor->setFrameStamp(_frameStamp.get());

    // use the frame number for the traversal number.
    if (_frameStamp.valid())
    {
         cullVisitor->setTraversalNumber(_frameStamp->getFrameNumber());
    }

    cullVisitor->inheritCullSettings(*this);

    cullVisitor->setClearNode(NULL); // reset earth sky on each frame.
    
    cullVisitor->setStateGraph(rendergraph);
    cullVisitor->setRenderStage(renderStage);

#if defined(OSG_RENDERINFO)
    cullVisitor->setRenderInfo( _renderInfo );
#else
    cullVisitor->setState( _state.get() );
#endif


    renderStage->reset();

    // comment out reset of rendergraph since clean is more efficient.
    //  rendergraph->reset();

    // use clean of the rendergraph rather than reset, as it is able to
    // reuse the structure on the rendergraph in the next frame. This
    // achieves a certain amount of frame cohereancy of memory allocation.
    rendergraph->clean();

    renderStage->setViewport(getViewport());
    //renderStage->setClearColor(getClearColor());

    switch(_lightingMode)
    {
    case(HEADLIGHT):
        if (_light.valid()) renderStage->addPositionedAttribute(NULL,_light.get());
        else osg::notify(osg::WARN)<<"Warning: no osg::Light attached to ogUtil::SceneView to provide head light.*/"<<std::endl;
        break;
    case(SKY_LIGHT):
        if (_light.valid()) renderStage->addPositionedAttribute(mv.get(),_light.get());
        else osg::notify(osg::WARN)<<"Warning: no osg::Light attached to ogUtil::SceneView to provide sky light.*/"<<std::endl;
        break;
    default:
        break;
    }            

    if (_globalStateSet.valid()) cullVisitor->pushStateSet(_globalStateSet.get());
    if (_localStateSet.valid()) cullVisitor->pushStateSet(_localStateSet.get());


    cullVisitor->pushViewport(getViewport());
    cullVisitor->pushProjectionMatrix(proj.get());
    cullVisitor->pushModelViewMatrix(mv.get());
    

    // traverse the scene graph to generate the rendergraph.
    for(unsigned int childNo=0;
        childNo<_rootNode->getNumChildren();
        ++childNo)
    {
        _rootNode->getChild(childNo)->accept(*cullVisitor);
    }

    cullVisitor->popModelViewMatrix();
    cullVisitor->popProjectionMatrix();
    cullVisitor->popViewport();

    if (_localStateSet.valid()) cullVisitor->popStateSet();
    if (_globalStateSet.valid()) cullVisitor->popStateSet();
    

    const osg::ClearNode* clearNode = cullVisitor->getClearNode();
    if (clearNode)
    {
        if (clearNode->getRequiresClear())
        {
            renderStage->setClearColor(clearNode->getClearColor());
            renderStage->setClearMask(clearNode->getClearMask());
        }
        else
        {
            // we have an earth sky implementation to do the work for use
            // so we don't need to clear.
            renderStage->setClearMask(0);
        }
    }

    renderStage->sort();

    // prune out any empty StateGraph children.
    // note, this would be not required if the rendergraph had been
    // reset at the start of each frame (see top of this method) but
    // a clean has been used instead to try to minimize the amount of
    // allocation and deleteing of the StateGraph nodes.
    rendergraph->prune();
}

osg::Viewport *OsgSceneHandler::getViewport()
{
    if( _globalStateSet.valid() )
        return dynamic_cast<osg::Viewport *>(_globalStateSet->getAttribute( osg::StateAttribute::VIEWPORT ));
    else
        return 0L;
}

bool OsgSceneHandler::setViewport( int x, int y, unsigned int width, unsigned int height )
{
    osg::Viewport *vp = getViewport();
    if( vp != 0L )
    {
        vp->setViewport( x, y, width, height );
    }
    else
    {
        if( !_globalStateSet.valid() )
            return false;
        else
            _globalStateSet->setAttribute( new osg::Viewport( x, y, width, height ));
    }
    return true;
}

void OsgSceneHandler::_init()
{
    if ( _rootNode.valid() && _initVisitor.valid())
    {
        _initVisitor->reset();
        _initVisitor->setFrameStamp(_frameStamp.get());

        osgUtil::GLObjectsVisitor* dlv = dynamic_cast<osgUtil::GLObjectsVisitor*>(_initVisitor.get());

        if (dlv) 
            dlv->setState( getState() );
#if 0
#if defined(OSG_RENDERINFO)
            dlv->setState(_renderInfo.getState());
#else
            dlv->setState(_state.get() );
#endif
#endif

        if (_frameStamp.valid())
        {
             _initVisitor->setTraversalNumber(_frameStamp->getFrameNumber());
        }

        _rootNode->accept(*_initVisitor.get());
    }

    _initCalled = true;
}


#if defined(OSG_RENDERINFO)
void OsgSceneHandler::RenderStage::drawImplementation(osg::RenderInfo& renderInfo, osgUtil::RenderLeaf*& previous)
{
    RenderBin::drawImplementation(renderInfo,previous);
    renderInfo.getState()->apply();
}
#else
void OsgSceneHandler::RenderStage::drawImplementation(osg::State& state, osgUtil::RenderLeaf*& previous)
{
    RenderBin::drawImplementation(state,previous);
    state.apply();
}
#endif

void OsgSceneHandler::toggleLighting()
{
    _lighting = !_lighting;
    if( _lighting )
        _globalStateSet->setMode( GL_LIGHTING, osg::StateAttribute::ON );
    else
        _globalStateSet->setMode( GL_LIGHTING, osg::StateAttribute::OFF | osg::StateAttribute::OVERRIDE );
}

void OsgSceneHandler::toggleTexturing()
{
    _texturing = !_texturing;
    if( _texturing )
    {
        for( int i = 0; i < 8; i++ )
        _globalStateSet->setTextureMode( i, GL_TEXTURE_2D, osg::StateAttribute::ON );
    }
    else
    {
        for( int i = 0; i < 8; i++ )
        _globalStateSet->setTextureMode( i, GL_TEXTURE_2D, osg::StateAttribute::OFF | osg::StateAttribute::OVERRIDE );
    }
}

void OsgSceneHandler::toggleWireframe()
{
    _polygonMode = 
        _polygonMode == osg::PolygonMode::FILL ? (_scribe ? osg::PolygonMode::FILL : osg::PolygonMode::LINE) :
        _polygonMode == osg::PolygonMode::LINE ? osg::PolygonMode::POINT :
        _polygonMode == osg::PolygonMode::POINT ? osg::PolygonMode::FILL : osg::PolygonMode::FILL;

    if( _polygonMode == osg::PolygonMode::FILL )
    {
        if( (_scribe = !_scribe ))
            _scribeSwitch->setAllChildrenOn();
        else
            _scribeSwitch->setAllChildrenOff();
    }

    osg::PolygonMode *pmode = dynamic_cast<osg::PolygonMode *>(_globalStateSet->getAttribute( osg::StateAttribute::POLYGONMODE ));
    if( pmode == 0L )
    {
        pmode = new osg::PolygonMode;
        _globalStateSet->setAttribute( pmode );
    }
    pmode->setMode( osg::PolygonMode::FRONT_AND_BACK, _polygonMode );
}


void OsgSceneHandler::toggleBackfaceCulling()
{
    _backfaceCulling = !_backfaceCulling;
    if( _backfaceCulling ) 
        _globalStateSet->setMode( GL_CULL_FACE, osg::StateAttribute::ON );
    else 
        _globalStateSet->setMode( GL_CULL_FACE, osg::StateAttribute::OFF | osg::StateAttribute::OVERRIDE );

}

void OsgSceneHandler::_initScribeSwitch()
{
    _scribeSwitch = new osg::Switch;
    _scribeSwitch->setAllChildrenOff();

    osg::StateSet *stateset = _scribeSwitch->getOrCreateStateSet();
    // Default Material
    osg::ref_ptr<osg::Material> material = new osg::Material;
    
    // Polygon Offset
    osg::ref_ptr<osg::PolygonOffset> polyoffset = new osg::PolygonOffset;
    polyoffset->setFactor(-1.0f);
    polyoffset->setUnits(-1.0f);
    
    // Polygon Mode (line)
    osg::ref_ptr<osg::PolygonMode> polymode = new osg::PolygonMode;
    polymode->setMode(osg::PolygonMode::FRONT_AND_BACK,osg::PolygonMode::LINE);
    
    stateset->setAttributeAndModes(material.get(),osg::StateAttribute::OVERRIDE|osg::StateAttribute::ON);
    stateset->setAttributeAndModes(polyoffset.get(),osg::StateAttribute::OVERRIDE|osg::StateAttribute::ON);
    stateset->setAttributeAndModes(polymode.get(),osg::StateAttribute::OVERRIDE|osg::StateAttribute::ON);
    stateset->setMode(GL_LIGHTING,osg::StateAttribute::OVERRIDE|osg::StateAttribute::OFF);
    stateset->setTextureMode(0,GL_TEXTURE_2D,osg::StateAttribute::OVERRIDE|osg::StateAttribute::OFF);

    if( _rootNode.valid() )
        _rootNode->addChild(_scribeSwitch.get() );
}

void OsgSceneHandler::snapshot( char *filename )
{
    osg::ref_ptr<osg::Image>image = new osg::Image;
    osg::Viewport *vp = getViewport();
    image->allocateImage( (int)vp->width(), (int)vp->height(), 1, GL_RGB, GL_UNSIGNED_BYTE);
    image->readPixels( 0, 0, (int)vp->width(), (int)vp->height(), GL_RGB, GL_UNSIGNED_BYTE);

    osgDB::writeImageFile( *(image.get()), filename );
}

osgUtil::UpdateVisitor *OsgSceneHandler::getUpdateVisitor()
{
    return dynamic_cast<osgUtil::UpdateVisitor *>(_updateVisitor.get());
}

osgUtil::CullVisitor *OsgSceneHandler::getCullVisitor()
{
    return _cullVisitor.get();
}

int OsgSceneHandler::_contextIDCounter = 0;

} // namespace Producer



// Environment variable help.
static osg::ApplicationUsageProxy OsgSceneHandler_e1(
    osg::ApplicationUsage::ENVIRONMENTAL_VARIABLE,
    "PRODUCER_CAMERA_BLOCK_ON_VSYNC",
    "After the CPU finishes generating the scene for one frame,"
    " block the CPU until the start of the next frame.");

static osg::ApplicationUsageProxy OsgSceneHandler_e2(
    osg::ApplicationUsage::ENVIRONMENTAL_VARIABLE,
    "PRODUCER_CAMERA_CONFIG_FILE <filename>",
    "Specify the default Producer camera config file to use with"
    " ProducerOSG-based applications.");

static osg::ApplicationUsageProxy OsgSceneHandler_e3(
    osg::ApplicationUsage::ENVIRONMENTAL_VARIABLE,
    "PRODUCER_FILE_PATH <path[:path..]>",
    "Give one or more directory paths to search when looking for"
    " Camera configuration and other files.");

static osg::ApplicationUsageProxy OsgSceneHandler_e4(
    osg::ApplicationUsage::ENVIRONMENTAL_VARIABLE,
    "PRODUCER_THREAD_MODEL <model>",
    "Give the threading model to use for the Cameras. Possible values:"
    " SingleThreaded, ThreadPerRenderSurface, ThreadPerCamera.");

static osg::ApplicationUsageProxy OsgSceneHandler_e5(
    osg::ApplicationUsage::ENVIRONMENTAL_VARIABLE,
    "PRODUCER_STACK_SIZE <size>",
    "Give the program stack size. This option is not normally used.");

static osg::ApplicationUsageProxy OsgSceneHandler_e6(
    osg::ApplicationUsage::ENVIRONMENTAL_VARIABLE,
    "PRODUCER_OVERRIDE_REDIRECT",
    "???");


